@@ -61,8 +61,8 @@ import java.net.InetAddress;
 import java.net.ServerSocket;
 import java.net.Socket;
 import java.net.SocketException;
+
 import org.apache.commons.net.io.SocketInputStream;
-import org.apache.commons.net.SocketClient;
 
 /***
  * RCommandClient is very similar to
@@ -118,7 +118,7 @@ import org.apache.commons.net.SocketClient;
  * <p>
  * <p>
  * @author Daniel F. Savarese
- * @see SocketClient
+ * @see org.apache.commons.net.SocketClient
  * @see RExecClient
  * @see RLoginClient
  ***/
