@@ -336,10 +336,88 @@ public class FTPClient extends FTP
         __passivePort = index;
     }
 
-
-    // null arg if no arg
-    protected Socket __openDataConnection(int command, String arg)
+    private boolean __storeFile(int command, String remote, InputStream local)
     throws IOException
+    {
+        OutputStream output;
+        Socket socket;
+
+        if ((socket = _openDataConnection_(command, remote)) == null)
+            return false;
+
+        // TODO: Buffer size may have to be adjustable in future to tune
+        // performance.
+        output = new BufferedOutputStream(socket.getOutputStream(),
+                                          Util.DEFAULT_COPY_BUFFER_SIZE);
+        if (__fileType == ASCII_FILE_TYPE)
+            output = new ToNetASCIIOutputStream(output);
+        // Treat everything else as binary for now
+        try
+        {
+            Util.copyStream(local, output);
+        }
+        catch (IOException e)
+        {
+            try
+            {
+                socket.close();
+            }
+            catch (IOException f)
+            {}
+            throw e;
+        }
+        output.close();
+        socket.close();
+        return completePendingCommand();
+    }
+
+    private OutputStream __storeFileStream(int command, String remote)
+    throws IOException
+    {
+        OutputStream output;
+        Socket socket;
+
+        if ((socket = _openDataConnection_(command, remote)) == null)
+            return null;
+
+        output = socket.getOutputStream();
+        if (__fileType == ASCII_FILE_TYPE) {
+          // We buffer ascii transfers because the buffering has to
+          // be interposed between ToNetASCIIOutputSream and the underlying
+          // socket output stream.  We don't buffer binary transfers
+          // because we don't want to impose a buffering policy on the
+          // programmer if possible.  Programmers can decide on their
+          // own if they want to wrap the SocketOutputStream we return
+          // for file types other than ASCII.
+          output = new BufferedOutputStream(output,
+                                            Util.DEFAULT_COPY_BUFFER_SIZE);
+          output = new ToNetASCIIOutputStream(output);
+          
+        }
+        return new org.apache.commons.net.io.SocketOutputStream(socket, output);
+    }
+
+
+    /***
+     * Establishes a data connection with the FTP server, returning
+     * a Socket for the connection if successful.  If a restart
+     * offset has been set with {@link #setRestartOffset(long)},
+     * a REST command is issued to the server with the offset as
+     * an argument before establishing the data connection.  Active
+     * mode connections also cause a local PORT command to be issued.
+     * <p>
+     * @param command  The text representation of the FTP command to send.
+     * @param args The arguments to the FTP command.  If this parameter is
+     *             set to null, then the command is sent with no argument.
+     * @return A Socket corresponding to the established data connection.
+     *         Null is returned if an FTP protocol error is reported at
+     *         any point during the establishment and initialization of
+     *         the connection.
+     * @exception IOException  If an I/O error occurs while either sending a
+     *      command to the server or receiving a reply from the server.
+     ***/
+    protected Socket _openDataConnection_(int command, String arg)
+      throws IOException
     {
         Socket socket;
 
@@ -389,6 +467,11 @@ public class FTPClient extends FTP
             __parsePassiveModeReply((String)_replyLines.elementAt(0));
 
             socket = _socketFactory_.createSocket(__passiveHost, __passivePort);
+            if ((__restartOffset > 0) && !restart(__restartOffset))
+            {
+                socket.close();
+                return null;
+            }
 
             if (!FTPReply.isPositivePreliminary(sendCommand(command, arg)))
             {
@@ -418,54 +501,6 @@ public class FTPClient extends FTP
     }
 
 
-    private boolean __storeFile(int command, String remote, InputStream local)
-    throws IOException
-    {
-        OutputStream output;
-        Socket socket;
-
-        if ((socket = __openDataConnection(command, remote)) == null)
-            return false;
-
-        output = new BufferedOutputStream(socket.getOutputStream());
-        if (__fileType == ASCII_FILE_TYPE)
-            output = new ToNetASCIIOutputStream(output);
-        // Treat everything else as binary for now
-        try
-        {
-            Util.copyStream(local, output);
-        }
-        catch (IOException e)
-        {
-            try
-            {
-                socket.close();
-            }
-            catch (IOException f)
-            {}
-            throw e;
-        }
-        output.close();
-        socket.close();
-        return completePendingCommand();
-    }
-
-    private OutputStream __storeFileStream(int command, String remote)
-    throws IOException
-    {
-        OutputStream output;
-        Socket socket;
-
-        if ((socket = __openDataConnection(command, remote)) == null)
-            return null;
-
-        output = socket.getOutputStream();
-        if (__fileType == ASCII_FILE_TYPE)
-            output = new ToNetASCIIOutputStream(output);
-        return new org.apache.commons.net.io.SocketOutputStream(socket, output);
-    }
-
-
     protected void _connectAction_() throws IOException
     {
         super._connectAction_();
@@ -734,7 +769,7 @@ public class FTPClient extends FTP
     {
         __dataConnectionMode = PASSIVE_LOCAL_DATA_CONNECTION_MODE;
         // These will be set when just before a data connection is opened
-        // in __openDataConnection()
+        // in _openDataConnection_()
         __passiveHost = null;
         __passivePort = -1;
     }
@@ -1190,12 +1225,15 @@ public class FTPClient extends FTP
         InputStream input;
         Socket socket;
 
-        if ((socket = __openDataConnection(FTPCommand.RETR, remote)) == null)
+        if ((socket = _openDataConnection_(FTPCommand.RETR, remote)) == null)
             return false;
 
-        input = new BufferedInputStream(socket.getInputStream());
+        // TODO: Buffer size may have to be adjustable in future to tune
+        // performance.
+        input = new BufferedInputStream(socket.getInputStream(),
+                                        Util.DEFAULT_COPY_BUFFER_SIZE);
         if (__fileType == ASCII_FILE_TYPE)
-            input = new FromNetASCIIInputStream(input);
+          input = new FromNetASCIIInputStream(input);
         // Treat everything else as binary for now
         try
         {
@@ -1244,12 +1282,22 @@ public class FTPClient extends FTP
         InputStream input;
         Socket socket;
 
-        if ((socket = __openDataConnection(FTPCommand.RETR, remote)) == null)
+        if ((socket = _openDataConnection_(FTPCommand.RETR, remote)) == null)
             return null;
 
         input = socket.getInputStream();
-        if (__fileType == ASCII_FILE_TYPE)
-            input = new FromNetASCIIInputStream(input);
+        if (__fileType == ASCII_FILE_TYPE) {
+          // We buffer ascii transfers because the buffering has to
+          // be interposed between FromNetASCIIOutputSream and the underlying
+          // socket input stream.  We don't buffer binary transfers
+          // because we don't want to impose a buffering policy on the
+          // programmer if possible.  Programmers can decide on their
+          // own if they want to wrap the SocketInputStream we return
+          // for file types other than ASCII.
+          input = new BufferedInputStream(input,
+                                          Util.DEFAULT_COPY_BUFFER_SIZE);
+          input = new FromNetASCIIInputStream(input);
+        }
         return new org.apache.commons.net.io.SocketInputStream(socket, input);
     }
 
@@ -1744,7 +1792,11 @@ public class FTPClient extends FTP
      ***/
     public String getSystemName() throws IOException
     {
-        if (syst() == FTPReply.NAME_SYSTEM_TYPE)
+      //if (syst() == FTPReply.NAME_SYSTEM_TYPE)
+      // Technically, we should expect a NAME_SYSTEM_TYPE response, but
+      // in practice FTP servers deviate, so we soften the condition to
+      // a positive completion.
+        if (FTPReply.isPositiveCompletion(syst()))
             return ((String)_replyLines.elementAt(0)).substring(4);
 
         return null;
@@ -1846,7 +1898,7 @@ public class FTPClient extends FTP
         BufferedReader reader;
         Vector results;
 
-        if ((socket = __openDataConnection(FTPCommand.NLST, pathname)) == null)
+        if ((socket = _openDataConnection_(FTPCommand.NLST, pathname)) == null)
             return null;
 
         reader =
@@ -1927,7 +1979,7 @@ public class FTPClient extends FTP
         Socket socket;
         FTPFile[] results;
 
-        if ((socket = __openDataConnection(FTPCommand.LIST, pathname)) == null)
+        if ((socket = _openDataConnection_(FTPCommand.LIST, pathname)) == null)
             return null;
 
         results = parser.parseFileList(socket.getInputStream());
@@ -2020,6 +2072,110 @@ public class FTPClient extends FTP
         return listFiles(__fileListParser);
     }
 
+    /**
+     * Using a programmer specified <code> FTPFileEntryParser </code>,
+     * initialize an object containing a raw file information for the
+     * current working directory.  This information is obtained through
+     * the LIST command.  This object is then capable of being iterated to
+     * return a sequence of FTPFile objects with information filled in by the
+     * <code> FTPFileEntryParser </code> used.
+     * <p>
+     * The server may or may not expand glob expressions.  You should avoid
+     * using glob expressions because the return format for glob listings
+     * differs from server to server and will likely cause this method to fail.
+     * <p>
+     * This method differs from using the listFiles() methods in that
+     * expensive FTPFile objects are not created until needed which may be
+     * an advantage on large lists.
+     * 
+     * @param parser The <code> FTPFileEntryParser </code> that should be
+     *               used to parse each server file listing.
+     * 
+     * @return An iteratable object that holds the raw information and is
+     *         capable of providing parsed FTPFile objects, one for each file containing
+     *         information contained in the given path in the format determined by the
+     *         <code> parser </code> parameter.   Null will be returned if a
+     *         data connection cannot be opened.  If the current working directory
+     *         contains no files, an empty array will be the return.
+     * @example <pre>
+     *    FTPClient f=FTPClient();
+     *    f.connect(server);
+     *    f.login(username, password);
+     *    FTPFileList list = createFTPFileList(directory, parser);
+     *    FTPFileIterator iter = list.iterator();
+     * 
+     *    while (iter.hasNext()) {
+     *       FTPFile[] files = iter.getNext(25);  // "page size" you want
+     *       //do whatever you want with these files, display them, etc.
+     *       //expensive FTPFile objects not created until needed.
+     *    }
+     * </pre>
+     * 
+     * @exception FTPConnectionClosedException
+     *                   If the FTP server prematurely closes the connection as a result
+     *                   of the client being idle or some other reason causing the server
+     *                   to send FTP reply code 421.  This exception may be caught either
+     *                   as an IOException or independently as itself.
+     * @exception IOException
+     *                   If an I/O error occurs while either sending a
+     *                   command to the server or receiving a reply from the server.
+     * @see FTPFileList
+     */
+    public FTPFileList createFileList(FTPFileEntryParser parser)
+    throws IOException
+    {
+        return createFileList(null, parser);
+    }
+
+    /**
+     * Using a programmer specified <code> FTPFileEntryParser </code>, 
+     * initialize an object containing a raw file information for a directory 
+     * or information for a single file.  This information is obtained through 
+     * the LIST command.  This object is then capable of being iterated to 
+     * return a sequence of FTPFile objects with information filled in by the
+     * <code> FTPFileEntryParser </code> used.
+     * The server may or may not expand glob expressions.  You should avoid
+     * using glob expressions because the return format for glob listings
+     * differs from server to server and will likely cause this method to fail.
+     * <p>
+     * @param parser The <code> FTPFileEntryParser </code> that should be
+     *         used to parse each server file listing.   
+     * @param pathname  The file or directory to list.
+     * @return An iteratable object that holds the raw information and is 
+     * capable of providing parsed FTPFile objects, one for each file containing
+     * information contained in the given path in the format determined by the 
+     * <code> parser </code> parameter.  Null will be returned if a 
+     * data connection cannot be opened.  If the supplied path contains
+     * no files, an empty array will be the return.
+     * @exception FTPConnectionClosedException
+     *      If the FTP server prematurely closes the connection as a result
+     *      of the client being idle or some other reason causing the server
+     *      to send FTP reply code 421.  This exception may be caught either
+     *      as an IOException or independently as itself.
+     * @exception IOException  If an I/O error occurs while either sending a
+     *      command to the server or receiving a reply from the server.
+     * @see FTPFileList
+     */
+    public FTPFileList createFileList(String pathname,
+                                      FTPFileEntryParser parser)
+    throws IOException
+    {
+        Socket socket;
+
+        if ((socket = _openDataConnection_(FTPCommand.LIST, pathname)) == null)
+        {
+            return null;
+        }
+
+        FTPFileList list =
+            FTPFileList.create(socket.getInputStream(), parser);
+
+        socket.close();
+
+        completePendingCommand();
+
+        return list;
+    }
 
     /***
      * Issue the FTP STAT command to the server.
